/*
Copyright (c) Andrew I MacIntyre 2008

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/*
exrtmodule.c

A Python interface to the EMX Runtime Library, providing access to those
non-portable, butstill useful routines.

Only ever compiled with gcc, so no attempt has been made to avoid EMX/GNU
extensions, etc...

This is expected to work only on OS/2 with EMX...

inspired by msvcrtmodule.c (for MSVC)
*/

#include "Python.h"
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <sys/ea.h>
#include <sys/ead.h>


/***************************************
 * OS/2 specific information reporting *
 ***************************************/

/* get full path name of executable file for current process */
static char emxrt_execname_doc[] =
"execname()\n"
"\n"
"Get the full pathname of the executable file for the current process\n";

static PyObject *
emxrt_execname(PyObject *self, PyObject *args)
{
	char s[512];

	if (!PyArg_ParseTuple(args, ":execname"))
		return NULL;

	if (_execname(s, 512) != 0)
		return PyErr_SetFromErrno(PyExc_IOError);

	return PyString_FromString(s);
}

/* get file system type of the specified drive */
static char emxrt_filesys_doc[] =
"filesys(drive)\n"
"\n"
"Get the file system type for the specified drive\n";

static PyObject *
emxrt_filesys(PyObject *self, PyObject *args)
{
	char *drv;
	char s[64];
	int rc;

	if (!PyArg_ParseTuple(args, "s:filesys", &drv))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	rc = _filesys(drv, s, 64);
	Py_END_ALLOW_THREADS
	if (rc != 0)
		return PyErr_SetFromErrno(PyExc_IOError);

	return PyString_FromString(s);
}

/****************************************
 * text/binary file translation control *
 ****************************************/

/* Set the file translation mode for a C runtime file descriptor */
static char emxrt_setmode_doc[] =
"setmode(fd, mode)\n"
"\n"
"Set the newline translation mode of file descriptor fd to mode\n"
"(mode is either O_TEXT or O_BINARY)\n";

static PyObject *
emxrt_setmode(PyObject *self, PyObject *args)
{
	int fd;
	int flags;

	if (!PyArg_ParseTuple(args,"ii:setmode", &fd, &flags))
		return NULL;

	flags = _setmode(fd, flags);
	if (flags == -1)
		return PyErr_SetFromErrno(PyExc_IOError);

	return PyInt_FromLong(flags);
}

/* Set the file translation mode for a C runtime file stream */
static char emxrt_fsetmode_doc[] =
"fsetmode(file, mode)\n"
"\n"
"Set the newline translation mode of a python file object\n"
"mode can be 't' (text) or 'b' (binary)\n";

static PyObject *
emxrt_fsetmode(PyObject *self, PyObject *args)
{
	PyObject *fobj;
	char fmode[2];
	FILE *fstr;

	if (!PyArg_ParseTuple(args,"Oc:setmode", &fobj, &fmode))
		return NULL;
	fmode[1] = '\0';

	/* did we get handed a file object to work with? */
	if (!PyFile_Check(fobj))
	{
		PyErr_SetString(PyExc_TypeError, "must be file object");
		return NULL;
	}
	fstr = PyFile_AsFile(fobj);

	/* sensible mode? */
	if (!(fmode[0] == 't' || fmode[0] == 'b'))
	{
		PyErr_SetString(PyExc_ValueError, "illegal mode");
		return NULL;
	}

	/* EMX requires that the stream be fflush()ed before a mode change */
	if (fflush(fstr) == EOF)
	{
		PyErr_SetString(PyExc_IOError, "write error");
		return NULL;
	}
	if (_fsetmode(fstr, fmode) != 0)
	{
		PyErr_SetString(PyExc_IOError, "mode change failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}

/*******************************
 * OS/2 file access management *
 *******************************/

/* open a file descriptor with sharing modes */
static char emxrt_sopen_doc[] =
"sopen(name, o_flags, sh_flags)\n"
"\n"
"Open a file descriptor for the named file\n"
"o_flags are the same as for os.open()\n"
"sh_flags can be one of SH_DENYRW, SH_DENYRD, SH_DENYWR, SH_DENYNO\n";

static PyObject *
emxrt_sopen(PyObject *self, PyObject *args)
{
	char* name;
	int o_flags;
	int sh_flags;
	int fd;

	if (!PyArg_ParseTuple(args, "sii:sopen", &name, &o_flags, &sh_flags))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	fd = sopen(name, o_flags, sh_flags);
	Py_END_ALLOW_THREADS
	if (fd == -1)
		return PyErr_SetFromErrnoWithFilename(PyExc_IOError, name);

	return PyInt_FromLong(fd);
}

/* create a Python file object with sharing modes */
static char emxrt_fsopen_doc[] =
"fsopen(name, mode, sh_flags)\n"
"\n"
"Create a file object for the named file\n"
"mode is the same as for open()\n"
"sh_flags can be one of SH_DENYRW, SH_DENYRD, SH_DENYWR, SH_DENYNO\n";

static PyObject *
emxrt_fsopen(PyObject *self, PyObject *args)
{
	char *name;
	char *mode;
	int sh_flags;
	FILE *f;

	if (!PyArg_ParseTuple(args,"ssi:fsopen", &name, &mode, &sh_flags))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	f = _fsopen(name, mode, sh_flags);
	Py_END_ALLOW_THREADS
	if (f == NULL)
		return PyErr_SetFromErrnoWithFilename(PyExc_IOError, name);

	return PyFile_FromFile(f, name, mode, fclose);
}

/*****************************************
 * OS/2 Extended Attribute (EA) routines *
 *****************************************/

/* return the value of a specified EA */
static char emxrt_ea_get_doc[] =
"ea_get(file, ea_name)\n"
"\n"
"Return a tuple (ea_data, ea_flags) for the specified OS/2 extended attribute.\n"
"file:     file object, file handle or file name string;\n"
"          a file object or handle requires an open mode including\n"
"          read access.\n"
"ea_name:  a string containing the name of the EA; case is ignored.\n"
"\n"
"The returned EA data is an arbitrary string value.\n"
"A KeyError exception is raised when the EA doesn't exist.\n";

static PyObject *
emxrt_ea_get(PyObject *self, PyObject *args)
{
	PyObject *file;
	PyObject *ea_ret;
	char *name;
	int fdesc = 0;
	char *fname = NULL;
	struct _ea ea_result;
	int rc;

	if (!PyArg_ParseTuple(args,"Os:ea_get", &file, &name))
		return NULL;

	/* did we get handed a file object to work with? */
	if (PyFile_Check(file))
	{
		fdesc = fileno(PyFile_AsFile(file));
	}
	else if (PyInt_Check(file))
	{
		fdesc = (int) PyInt_AsLong(file);
	}
	else if (PyString_Check(file))
	{
		fname = PyString_AsString(file);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError, "must be file object, file handle or string file name");
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	rc = _ea_get(&ea_result, fname, fdesc, name);
	Py_END_ALLOW_THREADS

	/* error getting EA */
	if (rc != 0)
	{
		if (fname != NULL)
			return PyErr_SetFromErrnoWithFilename(PyExc_IOError, fname);
		else
			return PyErr_SetFromErrno(PyExc_IOError);
	}

	/* EA doesn't exist */
	if (ea_result.value == NULL)
	{
		PyErr_SetString(PyExc_KeyError, "extended attribute doesn't exist");
		return NULL;
	}

	/* build the return tuple and free the EA value storage */
	ea_ret = Py_BuildValue("s#i", ea_result.value, ea_result.size, ea_result.flags);
	_ea_free(&ea_result);

	return ea_ret;
}

/* add/update an EA to a file */
static char emxrt_ea_put_doc[] =
"ea_put(file, ea_name, ea_value [, ea_flags])\n"
"\n"
"Add an extended attribute to a file or directory, returns None.\n"
"file:     may be a file object, a file handle or a file name string;\n"
"          a file object or handle requires an open mode including\n"
"          write access.\n"
"ea_name:  a string containing the name of the EA; case is ignored.\n"
"ea_value: the string to be set as the EA value.\n"
"ea_flags: an integer containing flag bits for the EA (default 0);\n"
"          - bit 7 (0x80) = critical EA.\n"
"\n"
"EMX documentation cautions:\n"
"\"flags should be 0 unless you know exactly what you are doing\"\n";

static PyObject *
emxrt_ea_put(PyObject *self, PyObject *args)
{
	PyObject *file;
	char *ea_name;
	struct _ea ea_src;
	char *fname = NULL;
	int fdesc = 0;
	int rc;

	ea_src.flags = 0;
	if (!PyArg_ParseTuple(args,"Oss#|i:ea_put",
	                      &file,
	                      &ea_name,
	                      &ea_src.value,
	                      &ea_src.size,
	                      &ea_src.flags))
		return NULL;

	/* did we get handed a file object to work with? */
	if (PyFile_Check(file))
	{
		fdesc = fileno(PyFile_AsFile(file));
	}
	else if (PyInt_Check(file))
	{
		fdesc = (int) PyInt_AsLong(file);
	}
	else if (PyString_Check(file))
	{
		fname = PyString_AsString(file);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
		                "must be file object, file handle or string file name");
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	rc = _ea_put(&ea_src, fname, fdesc, ea_name);
	Py_END_ALLOW_THREADS

	/* error setting EA */
	if (rc != 0)
	{
		if (fname != NULL)
			return PyErr_SetFromErrnoWithFilename(PyExc_IOError, fname);
		else
			return PyErr_SetFromErrno(PyExc_IOError);
	}

	/* apparent success, so return None */
	Py_INCREF(Py_None);
	return Py_None;
}

/* remove an EA from a file */
static char emxrt_ea_remove_doc[] =
"ea_remove(file, ea_name)\n"
"\n"
"Remove an extended attribute from a file or directory, returns None.\n"
"file:     may be a file object, a file handle or a file name string;\n"
"          a file object or handle requires an open mode including\n"
"          write access.\n"
"ea_name:  a string containing the name of the EA; case is ignored.\n";

static PyObject *
emxrt_ea_remove(PyObject *self, PyObject *args)
{
	PyObject *file;
	char *ea_name;
	char *fname = NULL;
	int fdesc = 0;
	int rc;

	if (!PyArg_ParseTuple(args,"Os:ea_remove", &file, &ea_name))
		return NULL;

	/* did we get handed a file object to work with? */
	if (PyFile_Check(file))
	{
		fdesc = fileno(PyFile_AsFile(file));
	}
	else if (PyInt_Check(file))
	{
		fdesc = (int) PyInt_AsLong(file);
	}
	else if (PyString_Check(file))
	{
		fname = PyString_AsString(file);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
		                "must be file object, file handle or string file name");
		return NULL;
	}

	Py_BEGIN_ALLOW_THREADS
	rc = _ea_remove(fname, fdesc, ea_name);
	Py_END_ALLOW_THREADS

	/* error removing EA */
	if (rc != 0)
	{
		if (fname != NULL)
			return PyErr_SetFromErrnoWithFilename(PyExc_IOError, fname);
		else
			return PyErr_SetFromErrno(PyExc_IOError);
	}

	/* apparent success, so return None */
	Py_INCREF(Py_None);
	return Py_None;
}

/* retrieve all EAs associated with a file */
static char emxrt_ea_getall_doc[] =
"ea_getall(file)\n"
"\n"
"Get all extended attributes associated with a file or directory,\n"
"returns a dictionary with the EA names as keys and item values being\n"
"tuples containing the EA values as strings and EA flags as integers.\n"
"file:     may be a file object, a file handle or a file name string;\n"
"          a file object or handle requires an open mode including\n"
"          read access.\n";

static PyObject *
emxrt_ea_getall(PyObject *self, PyObject *args)
{
	PyObject *file;
	char *fname = NULL;
	int fdesc = 0;
	_ead ea_desc;
	PyObject *ea_dict;
	int rc;
	int i, j;
	PyObject *k, *v, *f, *t;
	int s;

	if (!PyArg_ParseTuple(args,"O:ea_getall", &file))
		return NULL;

	/* did we get handed a file object to work with? */
	if (PyFile_Check(file))
	{
		fdesc = fileno(PyFile_AsFile(file));
	}
	else if (PyInt_Check(file))
	{
		fdesc = (int) PyInt_AsLong(file);
	}
	else if (PyString_Check(file))
	{
		fname = PyString_AsString(file);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
		                "must be file object, file handle or string file name");
		return NULL;
	}

	/* setup the EA descriptor structure */
	if ((ea_desc = _ead_create()) == NULL)
		return PyErr_SetFromErrno(PyExc_IOError);

	/* read all the EAs into the descriptor */
	Py_BEGIN_ALLOW_THREADS
	rc = _ead_read(ea_desc, fname, fdesc, 0);
	Py_END_ALLOW_THREADS

	/* error reading EAs, clean up and bail out */
	if (rc != 0)
	{
		_ead_destroy(ea_desc);

		if (fname != NULL)
			return PyErr_SetFromErrnoWithFilename(PyExc_IOError, fname);
		else
			return PyErr_SetFromErrno(PyExc_IOError);
	}

	/* create the dictionary */
	if ((ea_dict = PyDict_New()) == NULL)
	{
		/* oops... creating dict failed so clean up and bail out */
		_ead_destroy(ea_desc);

		PyErr_SetString(PyExc_MemoryError, "dict create failed");
		return NULL;
	}

	/* load the EA data into the dictionary */
	j = _ead_count(ea_desc);
	for (i = 1; i <= j; i++)
	{
		k = PyString_FromString(_ead_get_name(ea_desc, i));
		s = _ead_value_size(ea_desc, i);
		v = PyString_FromStringAndSize(_ead_get_value(ea_desc, i), s);
		f = PyInt_FromLong((long) _ead_get_flags(ea_desc, i));
		t = PyTuple_New(2);
		if (k != NULL && v != NULL && f != NULL && t != NULL);
		{
			if (PyTuple_SetItem(t, 0, v) == 0)
				Py_INCREF(v);
			if (PyTuple_SetItem(t, 1, f) == 0)
				Py_INCREF(f);
			rc = PyDict_SetItem(ea_dict, k, t);
		}
		if (k != NULL)
			Py_DECREF(k);
		if (t != NULL)
			Py_DECREF(t);
		if (v != NULL)
			Py_DECREF(v);
		if (f != NULL)
			Py_DECREF(f);
	}

	/* finished with the EA descriptor */
	_ead_destroy(ea_desc);

	/* if last SetItem() failed, abandon dictionary and bail out */
	if (rc != 0)
	{
		Py_DECREF(ea_dict);
		return NULL;
	}

	/* apparent success, so return dictionary */
	return ea_dict;
}

/* associate a collection of EAs with a file */
static char emxrt_ea_putall_doc[] =
"ea_putall(file, EAs[, merge])\n"
"\n"
"Associate a collection of extended attributes with a file or directory.\n"
"file:     may be a file object, a file handle or a file name string;\n"
"          a file object or handle requires an open mode including\n"
"          write access.\n"
"EAs:      a dictionary containing the EA collection - keys as EA names,\n"
"          values as tuples of EA value strings and EA flag integers.\n"
"          A ValueError exception will be raised if a dictionary entry\n"
"          item value is not a string, integer tuple.\n"
"merge:    an integer flag indicating whether the EAs in the supplied\n"
"          dictionary replace all EAs already associated with the file,\n"
"          or are merged with the file's existing EAs\n"
"          replace: 0 (default), merge: 1.\n";

static PyObject *
emxrt_ea_putall(PyObject *self, PyObject *args)
{
	PyObject *file;
	PyObject *ea_dict;
	int flag = 0;
	char *fname = NULL;
	int fdesc = 0;
	_ead ea_desc;
	PyObject *k, *v, *f, *t;
	char *ea_name = NULL;
	char *ea_val = NULL;
	int ea_size = 0;
	int ea_flag = 0;
	int i;
	int rc, ok;

	if (!PyArg_ParseTuple(args,"OO|i:ea_putall", &file, &ea_dict, &flag))
		return NULL;

	/* did we get handed a file object to work with? */
	if (PyFile_Check(file))
	{
		fdesc = fileno(PyFile_AsFile(file));
	}
	else if (PyInt_Check(file))
	{
		fdesc = (int) PyInt_AsLong(file);
	}
	else if (PyString_Check(file))
	{
		fname = PyString_AsString(file);
	}
	else
	{
		PyErr_SetString(PyExc_TypeError,
		                "must be file object, file handle or string file name");
		return NULL;
	}

	/* did we get a dictionary full of EA name, value pairs? */
	if (!PyDict_Check(ea_dict))
	{
		PyErr_SetString(PyExc_ValueError,
		                "extended attributes must be in a dictionary");
		return NULL;
	}

	/* check the choice of flag */
	if (flag != 0)
		flag = _EAD_MERGE;

	/* setup the EA descriptor structure */
	if ((ea_desc = _ead_create()) == NULL)
		return PyErr_SetFromErrno(PyExc_IOError);

	/* load the contents of the EA dictionary into the EA desriptor
	 * structure.  Raise a ValueError exception of the dictionary
	 * doesn't contain the appropriate data.
	 */
	i = 0;
	while (PyDict_Next(ea_dict, &i, &k, &t))
	{
		if (ok = PyString_Check(k))
			ea_name = PyString_AsString(k);

		if (ok)
		{
			if (PyList_Check(t) && (PyList_Size(t) >= 2))
			{
				v = PyList_GetItem(t, 0);
				f = PyList_GetItem(t, 1);
			}
			else if (PyTuple_Check(t) && (PyTuple_Size(t) >= 2))
			{
				v = PyTuple_GetItem(t, 0);
				f = PyTuple_GetItem(t, 1);
			}
			else
			{
				ok = 0;
			}
		}

		if (ok && (ok = PyString_Check(v)))
			ok = (PyString_AsStringAndSize(v, &ea_val, &ea_size) == 0);

		if (ok && (ok =	PyInt_Check(f)))
			ea_flag = PyInt_AsLong(f);

		/* add the EA to the descriptor structure */
		if (ok)
		{
			if (_ead_add(ea_desc, ea_name, ea_flag, ea_val, ea_size) < 0)
			{
				_ead_destroy(ea_desc);
				return PyErr_SetFromErrno(PyExc_IOError);
			}
		}
		else
		{
			_ead_destroy(ea_desc);
			PyErr_SetString(PyExc_ValueError,
			                "EA dict value must be string, integer tuple");
			return NULL;
		}
	}

	/* write the EAs to the file/directory */
	Py_BEGIN_ALLOW_THREADS
	rc = _ead_write(ea_desc, fname, fdesc, flag);
	Py_END_ALLOW_THREADS

	/* finished with the EA descriptor */
	_ead_destroy(ea_desc);

	/* error writing EAs, clean up and bail out */
	if (rc != 0)
	{
		if (fname != NULL)
			return PyErr_SetFromErrnoWithFilename(PyExc_IOError, fname);
		else
			return PyErr_SetFromErrno(PyExc_IOError);
	}

	/* apparent success */
	Py_INCREF(Py_None);
	return Py_None;
}


/********************
 * module utilities *
 ********************/

/* stuff values into a dictionary */
static void
insertint(PyObject *d, char *name, int value)
{
	PyObject *v = PyInt_FromLong((long) value);
	if (v == NULL) {
		/* Don't bother reporting this error */
		PyErr_Clear();
	}
	else {
		PyDict_SetItemString(d, name, v);
		Py_DECREF(v);
	}
}


/********************
 * module interface *
 ********************/

static char module_doc[] =
"This module provides access to EMX C library runtime routines for:\n"
" - changing file newline translation modes (setmode(), fsetmode())\n"
" - opening files with OS/2 specific file sharing modes (sopen(), fsopen())\n"
" - retrieving OS/2 specific information (execname(), filesys())\n"
" - extended attribute manipulation"
"   (ea_get(), ea_getall(), ea_pu(), ea_putall, ea_remove())";

/* List of functions exported by this module */
static struct PyMethodDef emxrt_functions[] = {
	{"execname",	emxrt_execname, METH_VARARGS, emxrt_execname_doc},
	{"filesys",     emxrt_filesys, METH_VARARGS, emxrt_filesys_doc},
	{"setmode",	emxrt_setmode, METH_VARARGS, emxrt_setmode_doc},
	{"fsetmode",	emxrt_fsetmode, METH_VARARGS, emxrt_fsetmode_doc},
	{"sopen",	emxrt_sopen, METH_VARARGS, emxrt_sopen_doc},
	{"fsopen",	emxrt_fsopen, METH_VARARGS, emxrt_fsopen_doc},
	{"ea_get",	emxrt_ea_get, METH_VARARGS, emxrt_ea_get_doc},
	{"ea_getall",	emxrt_ea_getall, METH_VARARGS, emxrt_ea_getall_doc},
	{"ea_put",	emxrt_ea_put, METH_VARARGS, emxrt_ea_put_doc},
	{"ea_putall",	emxrt_ea_putall, METH_VARARGS, emxrt_ea_putall_doc},
	{"ea_remove",	emxrt_ea_remove, METH_VARARGS, emxrt_ea_remove_doc},
	{NULL,		NULL, 0, NULL}
};


/* module initialisation */
void
initemxrt(void)
{
	PyObject *m = Py_InitModule3("emxrt", emxrt_functions, module_doc);
	PyObject *d = PyModule_GetDict(m);

	/* constants for the sopen() function's o_flags argument */
	insertint(d, "O_RDONLY", O_RDONLY);
	insertint(d, "O_WRONLY", O_WRONLY);
	insertint(d, "O_RDWR", O_RDWR);
	insertint(d, "O_APPEND", O_APPEND);
	insertint(d, "O_CREAT", O_CREAT);
	insertint(d, "O_TRUNC", O_TRUNC);
	insertint(d, "O_EXCL", O_EXCL);
	insertint(d, "O_NONBLOCK", O_NONBLOCK);	/* ignored */
	insertint(d, "O_BINARY", O_BINARY);
	insertint(d, "O_TEXT", O_TEXT);
	insertint(d, "O_NOINHERIT ", O_NOINHERIT);
	insertint(d, "O_SYNC", O_SYNC);
	/*insertint(d, "O_SIZE", O_SIZE);*/	/* EMX extension, disabled */

	/* sharing mode flags */
	insertint(d, "SH_DENYRW", SH_DENYRW);	/* deny read & write */
	insertint(d, "SH_DENYRD", SH_DENYRD);	/* deny read access */
	insertint(d, "SH_DENYWR", SH_DENYWR);	/* deny write access */
	insertint(d, "SH_DENYNO", SH_DENYNO);	/* deny nothing (permit rd/wr) */
}
